"""Create constant and point scatterer models."""
from timeit import repeat
import numpy as np
import pytest
from smii.modeling.propagators.propagators import Scalar2D as Scalar2Dcpu
from smii.modeling.propagators.scalar2d_gpu import Scalar2D as Scalar2Dgpu
from smii.modeling.wavelets.wavelets import ricker
from smii.modeling.forward_model import forward_model
from smii.test.models_scalar import direct_2d_approx

@pytest.fixture
def _versions(cpuprop_kwargs, gpuprop_kwargs):
    return [{'name': 'Scalar2Dcpu', 'propagator': Scalar2Dcpu,
             'kwargs': cpuprop_kwargs},
             {'name': 'Scalar2Dgpu', 'propagator': Scalar2Dgpu,
             'kwargs': gpuprop_kwargs}]

def _setup_propagator(c, freq, dx, dt, nt, nx, num_shots,
                      num_sources_per_shot, num_receivers_per_shot,
                      prop_kwargs):
    model = np.ones(nx, np.float32) * c

    x_s_idx = np.ones([num_shots, num_sources_per_shot, 2], np.int)
    x_s_idx[:, :, 1] = np.tile(np.linspace(1, nx[1]-1, num_sources_per_shot),
                               [num_shots, 1])
    f = ricker(freq, nt, dt, 0.05)

    source = {}
    source['amplitude'] = np.tile(f, [num_shots, num_sources_per_shot, 1])
    source['locations'] = x_s_idx

    x_r_idx = np.ones([num_shots, num_receivers_per_shot, 2], np.int)
    x_r_idx[:, :, 1] = np.tile(np.linspace(1, nx[1]-1, num_receivers_per_shot),
                               [num_shots, 1])
    x_r_idx[:, :, 0] = nx[0] - 1

    prop_args = (model, dx, dt, source)
    if len(prop_kwargs) == 0:
        prop_kwargs = {'pml_width': 30}

    return prop_args, prop_kwargs, x_r_idx


def verify_version(propagator, x_r_idx):
    dx = propagator.geometry.dx
    dt = propagator.timestep.dt * propagator.timestep.step_ratio
    c = propagator.model.property['vp'][0, 0]
    f = propagator.sources.amplitude[0, 0].ravel()
    expected = []
    for shotidx in range(propagator.sources.locations.shape[0]):
        expected.append([])
        for sourceidx in range(propagator.sources.locations.shape[1]):
            expected[-1].append([])
            x_s_idx = propagator.sources.locations[shotidx, sourceidx]
            for receiveridx in range(x_r_idx.shape[1]):
                x_r_idx0 = x_r_idx[shotidx, receiveridx]
                expected[-1][-1].append(direct_2d_approx(x_r_idx0*dx,
                                                         x_s_idx*dx,
                                                         dx, dt, c, f))
    expected = np.sum(np.array(expected), axis=1)
    actual, _ = forward_model(propagator, x_r_idx)
    #np.allclose(expected, actual, atol=1)
    return expected, actual


def time_version(propagator, prop_args, prop_kwargs, x_r_idx):
    def closure():
        """Closure over variables so they can be used in repeat below."""
        prop = propagator(*prop_args, **prop_kwargs)
        forward_model(prop, x_r_idx)

    return np.min(repeat(closure, number=1))


def run_timing(versions, c=1500, freq=25, dx=5, dt=0.005, nt=3000, nx=[64, 64],
               num_shots=3, num_sources_per_shot=4, num_receivers_per_shot=5):

    timing = []
    for version in versions:
        prop_args, prop_kwargs, x_r_idx = \
                _setup_propagator(c, freq, dx, dt, nt, nx,
                                  num_shots,
                                  num_sources_per_shot,
                                  num_receivers_per_shot,
                                  version['kwargs'])
        runtime = time_version(version['propagator'], prop_args, prop_kwargs,
                               x_r_idx)
        timing.append({'name': version['name'], 'time': runtime})
    print(timing)
    return timing


def run_verify(versions, c=1500, freq=25, dx=5, dt=0.005, nt=3000, nx=[64, 64],
               num_shots=3, num_sources_per_shot=4, num_receivers_per_shot=5):

    o = []
    for version in versions:
        prop_args, prop_kwargs, x_r_idx = \
                _setup_propagator(c, freq, dx, dt, nt, nx,
                                  num_shots,
                                  num_sources_per_shot,
                                  num_receivers_per_shot,
                                  version['kwargs'])
        propagator = version['propagator'](*prop_args, **prop_kwargs)
        o.append(verify_version(propagator, x_r_idx))
    return o


def test_verify(_versions):
    o = run_verify(versions=_versions)
    assert np.linalg.norm((o[0][0] - o[0][1].receivers).ravel()) < 2


if __name__ == '__main__':

    run_timing()



class StoreWavefield(object):
    def __init__(self, propagator):
        raise NotImplementedError

    def store(self, wavefield, step):
        raise NotImplementedError

    def restore(self, step):
        raise NotImplementedError



def backpropagate(propagator, image, stored_source_wavefield):

    for step in reversed(range(1, propagator.timestep.num_steps)):
        source_wavefield = stored_source_wavefield.restore(step-1)
        receiver_wavefield = propagator.step()
        image.add(source_wavefield, receiver_wavefield)

    return image.image



#!/usr/bin/env python
def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('wavelets', parent_package, top_path)
    return config

if __name__ == "__main__":
    from numpy.distutils.core import setup
    setup(configuration=configuration)



from smii.modeling.store_wavefield.in_memory import InMemory
from smii.modeling.record_receivers.nearest_cell import NearestCell

def forward_model(propagator, receiver_locations,
                  record_receivers=True,
                  store_wavefield=False):

        recorder = _record_receivers(propagator, receiver_locations,
                                     record_receivers)
        storer = _store_wavefield(propagator, store_wavefield)

        num_steps = propagator.timestep.num_steps
        for step in range(1, num_steps):

            wavefield = propagator.step()

            if record_receivers:
                recorder.record(wavefield, step)

            if store_wavefield:
                storer.store(wavefield, step)

        return recorder, storer


def _record_receivers(propagator, receiver_locations, record_receivers):

        if record_receivers is True:
            record_receivers = NearestCell

        if record_receivers:
            return record_receivers(propagator, receiver_locations)
 

def _store_wavefield(propagator, store_wavefield):

        if store_wavefield is True:
            store_wavefield = InMemory

        if store_wavefield:
            return store_wavefield(propagator)



import pycuda.autoinit
import pycuda.driver as drv
import numpy as np
from smii.modeling.propagators.propagators import ScalarPropagator


class PycudaPropagator(ScalarPropagator):
    """PyCUDA implementations."""
    def __init__(self, jitfunc1, jitfunc2, fd1_d, fd2_d, model, dx, source_dt,
                 sources, pad_width, pml_width=None):
        super(PycudaPropagator, self).__init__(model.astype(np.float32),
                                               np.float32(dx),
                                               np.float32(source_dt),
                                               sources,
                                               np.int32(pad_width),
                                               pml_width=pml_width)
        self.jitfunc1 = jitfunc1
        self.jitfunc2 = jitfunc2

        # allocate and copy model to GPU

        self.model.padded_property_gpu = {}
        self.model.padded_property_gpu['vp2dt2'] = \
                drv.mem_alloc(self.model.padded_property['vp2dt2'].nbytes)
        drv.memcpy_htod(self.model.padded_property_gpu['vp2dt2'],
                self.model.padded_property['vp2dt2'])

        # allocate and initialize wavefields
        self.wavefield.current_gpu = \
                drv.mem_alloc(self.wavefield.current.nbytes)
        drv.memset_d32(self.wavefield.current_gpu, 0,
                       self.wavefield.current.size)
        self.wavefield.previous_gpu = \
                drv.mem_alloc(self.wavefield.previous.nbytes)
        drv.memset_d32(self.wavefield.previous_gpu, 0,
                       self.wavefield.previous.size)

        # allocate and initialize PML arrays
        self.pml.sigma_gpu = []
        for dim in range(self.geometry.ndim):
            self.pml.phi[dim].current_gpu = \
                    drv.mem_alloc(self.pml.phi[dim].current.nbytes)
            drv.memset_d32(self.pml.phi[dim].current_gpu, 0,
                           self.pml.phi[dim].current.size)
            self.pml.phi[dim].previous_gpu = \
                    drv.mem_alloc(self.pml.phi[dim].previous.nbytes)
            drv.memset_d32(self.pml.phi[dim].previous_gpu, 0,
                           self.pml.phi[dim].previous.size)
            self.pml.sigma_gpu.append(drv.mem_alloc(self.pml.sigma[dim].nbytes))
            drv.memcpy_htod(self.pml.sigma_gpu[dim], self.pml.sigma[dim])

        # allocate and copy sources arrays
        self.sources.amplitude_gpu \
                = drv.mem_alloc(self.sources.amplitude.nbytes)
        drv.memcpy_htod(self.sources.amplitude_gpu,
                        self.sources.amplitude)

        self.sources.padded_locations_gpu \
                = drv.mem_alloc(self.sources.padded_locations.nbytes)
        drv.memcpy_htod(self.sources.padded_locations_gpu,
                        self.sources.padded_locations)


        # create and copy finite difference coeffs to constant memory
        self.fd1_d = fd1_d
        fd1 = np.array([8/12, -1/12], np.float32) / dx
        drv.memcpy_htod(self.fd1_d, fd1)

        self.fd2_d = fd2_d
        if self.geometry.ndim == 1:
            fd2 = np.array([-5/2, 4/3, -1/12], np.float32) / dx**2
        elif self.geometry.ndim == 2:
            fd2 = np.array([-10/2, 4/3, -1/12], np.float32) / dx**2
        drv.memcpy_htod(self.fd2_d, fd2)

        # set block and grid dimensions
        threadsperblockx = 32
        blockspergridx = ((self.geometry.propagation_shape_padded[-1]
                           + (threadsperblockx - 1))
                          // threadsperblockx)
        if self.geometry.ndim == 1:
            threadsperblockz = 1
            blockspergridz = self.sources.num_shots
        elif self.geometry.ndim == 2:
            threadsperblockz = 32
            blockspergridz = ((self.geometry.propagation_shape_padded[-2]
                               + (threadsperblockz - 1))
                              // threadsperblockz) * self.sources.num_shots

        self.griddim = int(blockspergridx), int(blockspergridz)
        self.blockdim = int(threadsperblockx), int(threadsperblockz), 1

    def finalise(self):
        del (self.model.padded_property_gpu['vp2dt2'],
             self.wavefield.current_gpu,
             self.wavefield.previous_gpu,
             self.sources.amplitude_gpu, self.sources.padded_locations_gpu,
             self.jitfunc1, self.jitfunc2)

        for dim in range(self.geometry.ndim):
            del(self.pml.sigma_gpu[dim], self.pml.phi[dim].current_gpu,
                self.pml.phi[dim].previous_gpu)



"""Create constant and point scatterer models."""
import numpy as np
from smii.test.models_scalar import (model_direct_1d, model_direct_2d,
                                     model_scatter_1d, model_scatter_2d)
from smii.modeling.propagators.propagators import Scalar1D as Scalar1Dcpu
from smii.modeling.propagators.propagators import Scalar2D as Scalar2Dcpu
from smii.modeling.propagators.scalar1d_gpu import Scalar1D as Scalar1Dgpu
from smii.modeling.propagators.scalar2d_gpu import Scalar2D as Scalar2Dgpu

def test_direct_1d_cpu(cpuprop_kwargs):
    expected, actual = model_direct_1d(propagator=Scalar1Dcpu,
                                       prop_kwargs=cpuprop_kwargs)
    assert np.linalg.norm(expected - actual) < 14.5


def test_direct_1d_gpu(gpuprop_kwargs):
    expected, actual = model_direct_1d(propagator=Scalar1Dgpu,
                                       prop_kwargs=gpuprop_kwargs)
    assert np.linalg.norm(expected - actual) < 14.5


def test_direct_2d_cpu(cpuprop_kwargs):
    expected, actual = model_direct_2d(propagator=Scalar2Dcpu,
                                       prop_kwargs=cpuprop_kwargs)
    assert np.linalg.norm(expected - actual) < 0.6


def test_direct_2d_gpu(gpuprop_kwargs):
    expected, actual = model_direct_2d(propagator=Scalar2Dgpu,
                                       prop_kwargs=gpuprop_kwargs)
    assert np.linalg.norm(expected - actual) < 0.6


def test_scatter_1d_cpu(cpuprop_kwargs):
    expected, actual = model_scatter_1d(propagator=Scalar1Dcpu,
                                        prop_kwargs=cpuprop_kwargs)
    t = 2000
    assert np.linalg.norm(expected[t:] - actual[t:]) < 1.1


def test_scatter_1d_gpu(gpuprop_kwargs):
    expected, actual = model_scatter_1d(propagator=Scalar1Dgpu,
                                        prop_kwargs=gpuprop_kwargs)
    t = 2000
    assert np.linalg.norm(expected[t:] - actual[t:]) < 1.1


def test_scatter_2d_cpu(cpuprop_kwargs):
    expected, actual = model_scatter_2d(propagator=Scalar2Dcpu,
                                        prop_kwargs=cpuprop_kwargs)
    t = 2000
    assert np.linalg.norm(expected[t:] - actual[t:]) < 0.081


def test_scatter_2d_gpu(gpuprop_kwargs):
    expected, actual = model_scatter_2d(propagator=Scalar2Dgpu,
                                        prop_kwargs=gpuprop_kwargs)
    t = 2000
    assert np.linalg.norm(expected[t:] - actual[t:]) < 0.081



import numpy as np

def ricker(freq, length, dt, peak_time):
    """Return a Ricker wavelet with the specified central frequency."""
    t = np.arange(length, dtype=np.float32) * dt - peak_time
    y = (1 - 2 * np.pi**2 * freq**2 * t**2) \
            * np.exp(-np.pi**2 * freq**2 * t**2)
    return y



import numpy as np

class ImagingCondition(object):
    def __init__(self, shape, dt):
        self.image = np.zeros(shape, np.float32)
        self.dt = dt

    def add(self, source_wavefield, receiver_wavefield):
        raise NotImplementedError



from smii.modeling.imaging_condition.imaging_condition import ImagingCondition

class ZeroLagXcor(ImagingCondition):
    def __init__(self, shape, dt):
        super(ZeroLagXcor, self).__init__(shape, dt)

    def add(self, source_wavefield, receiver_wavefield):
        self.image += source_wavefield * receiver_wavefield * self.dt



import numpy as np
from smii.modeling.store_wavefield.store_wavefield import StoreWavefield

class InMemory(StoreWavefield):
    def __init__(self, propagator):
        self.wavefield = np.zeros([propagator.timestep.num_steps] +
                                  propagator.geometry.propagation_shape,
                                  np.float32)

    def store(self, wavefield, step):
        self.wavefield[step] = wavefield

    def restore(self, step):
        return self.wavefield[step]



import json
import pytest

def pytest_addoption(parser):
    parser.addoption("--cpuprop_kwargs", action="store", default="{}",
                     help="options for CPU propagators as JSON")
    parser.addoption("--gpuprop_kwargs", action="store", default="{}",
                     help="options for GPU propagators as JSON")

@pytest.fixture
def cpuprop_kwargs(request):
    return json.loads(request.config.getoption("--cpuprop_kwargs"))

@pytest.fixture
def gpuprop_kwargs(request):
    return json.loads(request.config.getoption("--gpuprop_kwargs"))



"""Create constant and point scatterer models."""
import numpy as np
import scipy.special
import scipy.integrate
from scipy.ndimage.interpolation import shift
from smii.modeling.propagators.propagators import (Scalar1D, Scalar2D)
from smii.modeling.wavelets.wavelets import ricker
from smii.modeling.forward_model import forward_model
from smii.inversion.fwi import costjac

def direct_1d(x, x_s, dx, dt, c, f):
    """Use the 1D Green's function to determine the wavefield at a given
    location and time due to the given source.
    """
    r = np.abs(x - x_s)
    t_shift = (r/c) / dt + 1
    u = dx * dt * c / 2 * np.cumsum(shift(f, t_shift))
    return u


def direct_2d(x, t, x_s, dx, dt, c, f):
    """Use the 2D Green's function to determine the wavefield at a given
    location and time due to the given source.
    """
    r = np.linalg.norm(x - x_s)
    t_max = np.maximum(0, int((t - r/c) / dt))
    tmtp = t - np.arange(t_max) * dt
    summation = np.sum(f[:t_max] / np.sqrt(c**2 * tmtp**2 - r**2))
    u = dx**2 * dt * c / 2 / np.pi * summation
    return u


def direct_2d2(x, x_s, dx, dt, c, f):
    """Use the 2D Green's function to determine the wavefield at a given
    location and time due to the given source.
    """
    r = np.linalg.norm(x - x_s)
    nt = len(f)
    def func(tp, t):
        return f[int(tp / dt)] / np.sqrt(c**2 * (t - tp)**2 - r**2)
    u = np.zeros_like(f)
    t_max = int(r/c / dt)
    for t_idx in range(t_max):
        t = t_idx * dt
        u[t_idx] = scipy.integrate.quad(func, 0, t, (t+dt))[0]
    u *= dx**2 * dt * c / 2 / np.pi
    return u

def direct_2d_approx(x, x_s, dx, dt, c, f):
    """Same as direct_2d, but using an approximation to calculate the result
    for the whole time range of the source.
    """
    r = np.linalg.norm(x - x_s)
    nt = len(f)
    w = np.fft.rfftfreq(nt, dt)
    fw = np.fft.rfft(f)
    G = 1j / 4 * scipy.special.hankel1(0, -2 * np.pi * w * r / c)
    G[0] = 0
    s = G * fw * dx**2
    u = np.fft.irfft(s, nt)
    return u


def direct_3d(x, x_s, dx, dt, c, f):
    """Use the 3D Green's function to determine the wavefield at a given
    location and time due to the given source.
    """
    r = np.linalg.norm(x - x_s)
    t_shift = (r/c) / dt + 1
    u = dx**3 * dt / 4 / np.pi / r * shift(f, t_shift)
    return u


def scattered_1d(x, x_s, x_p, dx, dt, c, dc, f):
    u_p = direct_1d(x_p, x_s, dx, dt, c, f)
    du_pdt2 = np.gradient(np.gradient(u_p)) / dt**2
    u = 2 * dc / c**3 * direct_1d(x, x_p, dx, dt, c, du_pdt2)
    return u


def scattered_2d(x, x_s, x_p, dx, dt, c, dc, f):
    u_p = direct_2d_approx(x_p, x_s, dx, dt, c, f)
    du_pdt2 = np.gradient(np.gradient(u_p)) / dt**2
    u = 2 * dc / c**3 * direct_2d_approx(x, x_p, dx, dt, c, du_pdt2)
    return u


def scattered_3d(x, x_s, x_p, dx, dt, c, dc, f):
    u_p = direct_3d(x_p, x_s, dx, dt, c, f)
    du_sdt2 = np.gradient(np.gradient(u_p)) / dt**2
    u = 2 * dc / c**3 * direct_3d(x, x_p, dx, dt, c, du_pdt2)
    return u


def grad_1d(nx, x_r, x_s, x_p, dx, dt, c, dc, f):
    d = -scattered_1d(x_r, x_s, x_p, dx, dt, c, dc, f)[::-1]
    grad = np.zeros(nx, np.float32)
    for x_idx in range(nx):
        x = x_idx*dx
        u_r = direct_1d(x, x_r, dx, dt, c, d)[::-1]
        u_0 = direct_1d(x, x_s, dx, dt, c, f)
        du_0dt2 = np.gradient(np.gradient(u_0)) / dt**2
        grad[x_idx] = 2 * dt / c**3 * np.sum(u_r * du_0dt2)
    return grad


def grad_2d(nx, x_r, x_s, x_p, dx, dt, c, dc, f):
    d = -scattered_2d(x_r, x_s, x_p, dx, dt, c, dc, f)[::-1]
    grad = np.zeros(nx, np.float32)
    for z_idx in range(nx[0]):
        for x_idx in range(nx[1]):
            x = np.array([z_idx*dx, x_idx*dx])
            u_r = direct_2d_approx(x, x_r, dx, dt, c, d)[::-1]
            u_0 = direct_2d_approx(x, x_s, dx, dt, c, f)
            du_0dt2 = np.gradient(np.gradient(u_0)) / dt**2
            grad[z_idx, x_idx] = 2 * dt / c**3 * np.sum(u_r * du_0dt2)
    return grad


def grad_1d_fd(model_true, model_init, x_r, x_s, dx, dt, dc, f,
               propagator=None, prop_kwargs=None):
    x_r_idx, x_s_idx = (np.array([x_r, x_s]) / dx).astype(np.int)
    source, receiver_locations = _make_source_receiver(x_s_idx, x_r_idx, f)
    if propagator is None:
        propagator = Scalar1D
    if prop_kwargs is None:
        prop_kwargs = {}
    prop = propagator(model_true, dx, dt, source, **prop_kwargs)
    true_data, _ = forward_model(prop, receiver_locations)
    receiver = {}
    receiver['amplitude'] = true_data.receivers
    receiver['locations'] = receiver_locations
    dataset = [(source, receiver)]
    init_cost, fwi_grad = costjac(model_init, dataset, dx, dt, propagator,
                                  model_init.shape, compute_grad=True,
                                  prop_kwargs=prop_kwargs)

    nx = len(model_true)
    true_grad = np.zeros(nx, np.float32)
    for x_idx in range(nx):
        tmp_model = model_init.copy()
        tmp_model[x_idx] += dc
        new_cost, _ = costjac(tmp_model, dataset, dx, dt, propagator,
                              model_init.shape, compute_grad=False,
                              prop_kwargs=prop_kwargs)
        true_grad[x_idx] = (new_cost - init_cost) / dc
    return fwi_grad, true_grad


def grad_2d_fd(model_true, model_init, x_r, x_s, dx, dt, dc, f,
               propagator=None, prop_kwargs=None):
    x_r_idx, x_s_idx = (np.array([x_r, x_s]) / dx).astype(np.int)
    source, receiver_locations = _make_source_receiver(x_s_idx, x_r_idx, f)
    if propagator is None:
        propagator = Scalar2D
    if prop_kwargs is None:
        prop_kwargs = {}
    prop = propagator(model_true, dx, dt, source, **prop_kwargs)
    true_data, _ = forward_model(propagator, receiver_locations)
    receiver = {}
    receiver['amplitude'] = true_data.receivers
    receiver['locations'] = receiver_locations
    dataset = [(source, receiver)]
    init_cost, fwi_grad = costjac(model_init, dataset, dx, dt, propagator,
                                  model_init.shape, compute_grad=True,
                                  prop_kwargs=prop_kwargs)

    true_grad = np.zeros_like(model_true)
    for z_idx in range(model_true.shape[0]):
        for x_idx in range(model_true.shape[1]):
            tmp_model = model_init.copy()
            tmp_model[z_idx, x_idx] += dc
            new_cost, _ = costjac(tmp_model, dataset, dx, dt, propagator,
                                  model_init.shape, compute_grad=False,
                                  prop_kwargs=prop_kwargs)
            true_grad[z_idx, x_idx] = (new_cost - init_cost) / dc
    return fwi_grad, true_grad


def _make_source_receiver(x_s_idx, x_r_idx, f):
    source = {}
    source['amplitude'] = f.reshape(1, 1, -1)
    source['locations'] = x_s_idx.reshape(1, 1, -1)
    receiver_locations = x_r_idx.reshape(1, 1, -1)
    return source, receiver_locations


def _set_coords(x, dx):
    x_m = np.array(x) * dx
    x_idx = np.array(x)
    return x_m, x_idx

def model_direct_1d(c=1500, freq=25, dx=5, dt=0.0001, nx=80,
                    propagator=None, prop_kwargs=None):
    """Create a constant model, and the expected waveform at point,
       and the forward propagated wave.
    """
    model = np.ones(nx, dtype=np.float32) * c

    nt = int(2*nx*dx/c/dt)
    x_s, x_s_idx = _set_coords([[1]], dx)
    x_r, x_r_idx = _set_coords([[nx-1]], dx)
    f = ricker(freq, nt, dt, 0.05)

    expected = direct_1d(x_r, x_s, dx, dt, c, f)

    source, receiver_locations = _make_source_receiver(x_s_idx, x_r_idx, f)
    if propagator is None:
        propagator = Scalar1D
    if prop_kwargs is None:
        prop_kwargs = {}
    prop = propagator(model, dx, dt, source, **prop_kwargs)

    actual, _ = forward_model(prop, receiver_locations)

    return expected, actual.receivers.ravel()


def model_direct_2d(c=1500, freq=25, dx=5, dt=0.0001, nx=[50, 50],
                    propagator=None, prop_kwargs=None):
    """Create a constant model, and the expected waveform at point,
       and the forward propagated wave.
    """
    model = np.ones(nx, dtype=np.float32) * c

    nt = int(2*nx[0]*dx/c/dt)
    middle = int(nx[1]/2)
    x_s, x_s_idx = _set_coords([[1, middle]], dx)
    x_r, x_r_idx = _set_coords([[nx[0]-1, middle]], dx)
    #x_r, x_r_idx = _set_coords([[1, middle]], dx)
    f = ricker(freq, nt, dt, 0.05)

    expected = direct_2d_approx(x_r, x_s, dx, dt, c, f)

    source, receiver_locations = _make_source_receiver(x_s_idx, x_r_idx, f)
    if propagator is None:
        propagator = Scalar2D
    if prop_kwargs is None:
        prop_kwargs = {}
    prop = propagator(model, dx, dt, source, **prop_kwargs)

    actual, _ = forward_model(prop, receiver_locations)

    return expected, actual.receivers.ravel()


def model_scatter_1d(c=1500, dc=50, freq=25, dx=5, dt=0.0001, nx=100,
                     propagator=None, prop_kwargs=None):
    """Create a point scatterer model, and the expected waveform at point,
       and the forward propagated wave.
    """
    model = np.ones(nx, dtype=np.float32) * c

    nt = int((3*nx*dx/c + 0.05)/dt)
    x_s, x_s_idx = _set_coords([[1]], dx)
    x_r, x_r_idx = _set_coords([[1]], dx)
    x_p, x_p_idx = _set_coords([[nx-20]], dx)
    f = ricker(freq, nt, dt, 0.05)

    model[x_p_idx] += dc

    expected = scattered_1d(x_r, x_s, x_p, dx, dt, c, dc, f)

    source, receiver_locations = _make_source_receiver(x_s_idx, x_r_idx, f)
    if propagator is None:
        propagator = Scalar1D
    if prop_kwargs is None:
        prop_kwargs = {}
    prop = propagator(model, dx, dt, source, **prop_kwargs)

    actual, _ = forward_model(prop, receiver_locations)

    return expected, actual.receivers.ravel()


def model_scatter_2d(c=1500, dc=150, freq=25, dx=5, dt=0.0001, nx=[50, 50],
                     propagator=None, prop_kwargs=None):
    """Create a point scatterer model, and the expected waveform at point,
       and the forward propagated wave.
    """
    nx = np.array(nx)
    model = np.ones(nx, dtype=np.float32) * c

    nt = int((3*nx[0]*dx/c + 0.05)/dt)
    middle = int(nx[1]/2)
    x_s, x_s_idx = _set_coords([[1, middle]], dx)
    x_r, x_r_idx = _set_coords([[1, middle]], dx)
    x_p, x_p_idx = _set_coords([[nx[0]-10, middle]], dx)
    f = ricker(freq, nt, dt, 0.05)

    model[x_p_idx[0, 0], x_p_idx[0, 1]] += dc

    expected = scattered_2d(x_r, x_s, x_p, dx, dt, c, dc, f)

    source, receiver_locations = _make_source_receiver(x_s_idx, x_r_idx, f)
    if propagator is None:
        propagator = Scalar2D
    if prop_kwargs is None:
        prop_kwargs = {}
    prop = propagator(model, dx, dt, source, **prop_kwargs)

    actual, _ = forward_model(prop, receiver_locations)

    return expected, actual.receivers.ravel()


def model_grad_const_1d(c=1500, dc=1, freq=25, dx=5, dt=0.0001, nx=100,
                        propagator=None, prop_kwargs=None):
    """Create a point scatterer model, and the gradient.
    """

    nt = int((3*nx*dx/c + 0.1)/dt)
    x_s, x_s_idx = _set_coords([[1]], dx)
    x_r, x_r_idx = _set_coords([[1]], dx)
    x_p, x_p_idx = _set_coords([[nx-20]], dx)
    f = ricker(freq, nt, dt, 0.05)

    model_init = np.ones(nx, dtype=np.float32) * c
    model_true = model_init.copy()
    model_true[x_p_idx] += dc

    expected = grad_1d(nx, x_r, x_s, x_p, dx, dt, c, dc, f)
    fwi_grad, true_grad = grad_1d_fd(model_true, model_init, x_r, x_s, dx, dt,
                                     dc, f, propagator, prop_kwargs)

    return expected, fwi_grad, true_grad


def model_grad_const_2d(c=1500, dc=1, freq=25, dx=5, dt=0.0001, nx=[20, 20],
                        propagator=None, prop_kwargs=None):
    """Create a point scatterer model, and the gradient.
    """

    nt = int((3*nx[0]*dx/c + 0.1)/dt)
    middle = int(nx[1]/2)
    x_s, x_s_idx = _set_coords([[1, middle]], dx)
    x_r, x_r_idx = _set_coords([[1, middle]], dx)
    x_p, x_p_idx = _set_coords([[nx[0]-5, middle]], dx)
    f = ricker(freq, nt, dt, 0.05)

    model_init = np.ones(nx, dtype=np.float32) * c
    model_true = model_init.copy()
    model_true[x_p_idx[0, 0], x_p_idx[0, 1]] += dc

    expected = grad_2d(nx, x_r, x_s, x_p, dx, dt, c, dc, f)
    fwi_grad, true_grad = grad_2d_fd(model_true, model_init, x_r, x_s, dx, dt,
                                     dc, f, propagator, prop_kwargs)

    return expected, fwi_grad, true_grad


def model_grad_rand_1d(c=2000, randc=100, dc=1, freq=25, dx=5, dt=0.0001,
                       nx=100, propagator=None, prop_kwargs=None):
    """Create a point scatterer model, and the gradient.
    """

    nt = int((3*nx*dx/c + 0.1)/dt)
    x_s, x_s_idx = _set_coords([[1]], dx)
    x_r, x_r_idx = _set_coords([[1]], dx)
    x_p, x_p_idx = _set_coords([[nx-20]], dx)
    f = ricker(freq, nt, dt, 0.05)

    model_init = (np.random.rand(nx).astype(np.float32) * randc) + c
    model_true = model_init.copy()
    model_true += np.random.rand(nx).astype(np.float32) * dc

    fwi_grad, true_grad = grad_1d_fd(model_true, model_init, x_r, x_s, dx, dt,
                                     dc, f, propagator, prop_kwargs)

    return fwi_grad, true_grad


def model_grad_rand_2d(c=2000, randc=100, dc=1, freq=25, dx=5, dt=0.0001,
                       nx=[20, 20], propagator=None, prop_kwargs=None):
    """Create a point scatterer model, and the gradient.
    """

    nt = int((3*nx[0]*dx/c + 0.1)/dt)
    middle = int(nx[1]/2)
    x_s, x_s_idx = _set_coords([[1, middle]], dx)
    x_r, x_r_idx = _set_coords([[1, middle]], dx)
    x_p, x_p_idx = _set_coords([[nx[0]-5, middle]], dx)
    f = ricker(freq, nt, dt, 0.05)

    model_init = (np.random.rand(nx[0], nx[1]).astype(np.float32) * randc) + c
    model_true = model_init.copy()
    model_true += np.random.rand(nx[0], nx[1]).astype(np.float32) * dc

    fwi_grad, true_grad = grad_2d_fd(model_true, model_init, x_r, x_s, dx, dt,
                                     dc, f, propagator, prop_kwargs)

    return fwi_grad, true_grad






#!/usr/bin/env python
def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('smii', parent_package, top_path)
    config.add_subpackage('modeling')
    config.add_subpackage('imaging')
    config.add_subpackage('inversion')
    config.add_data_dir('test')
    #config.make_config_py()
    return config

if __name__ == "__main__":
    print('This is the wrong setup.py file to run')



import pycuda.driver as drv
from pycuda.compiler import SourceModule
import numpy as np
from smii.modeling.propagators.pycuda import PycudaPropagator

class Scalar1D(PycudaPropagator):
    def __init__(self, model, dx, source_dt, sources, pml_width=None,
                 nvcc_options=None):

        source = """
__constant__ float fd1_d[2];
__constant__ float fd2_d[3];

__global__ void step_d(const float *const wfc,
                float *wfp,
                const float *const phix,
                float *phixp,
                const float *const sigmax,
                const float *const model2_dt2,
                const float dt,
                const int nx)
{
        int x = blockIdx.x * blockDim.x + threadIdx.x;
        int b = blockIdx.y;
        int bx = b * nx + x;
        float wfc_xx;
        float wfc_x;
        float phix_x;
        bool in_domain = (x > 1) && (x < nx - 2);

        if (in_domain)
        {
                wfc_xx = (fd2_d[0] * wfc[bx] +
                                fd2_d[1] *
                                (wfc[bx + 1] +
                                 wfc[bx - 1]) +
                                fd2_d[2] *
                                (wfc[bx + 2] +
                                 wfc[bx - 2]));

                wfc_x = (fd1_d[0] *
                                (wfc[bx + 1] -
                                 wfc[bx - 1]) + 
                                 fd1_d[1] * 
                                 (wfc[bx + 2] -
                                 wfc[bx - 2]));

                phix_x = (fd1_d[0] *
                                (phix[bx + 1] -
                                 phix[bx - 1]) + 
                                 fd1_d[1] * 
                                 (phix[bx + 2] -
                                 phix[bx - 2]));

                wfp[bx] = 1 / (1 + dt * sigmax[x] / 2) *
                        (model2_dt2[x] *
                        (wfc_xx + phix_x) +
                        dt * sigmax[x] * wfp[bx] / 2 +
                        (2 * wfc[bx] - wfp[bx]));

                phixp[bx] = phix[bx] -
                        dt * sigmax[x] * (wfc_x + phix[bx]);

        }
}

__global__ void add_sources_d(float *wfp,
                const float *const model2_dt2,
                const float *const source_amplitude,
                const int *const sources_x,
                const int step,
                const int nx,
                const int nt, const int ns)
{
        int s = threadIdx.x;
        int b = blockIdx.x;
        int x = sources_x[b * ns + s];
        int bx = b * nx + x;

        wfp[bx] += source_amplitude[b * ns * nt + s * nt + step] *
            model2_dt2[x];
}
"""

        if nvcc_options is None:
            nvcc_options = ['--restrict', '--use_fast_math', '-O3']

        mod = SourceModule(source, options=nvcc_options)

        jitfunc1 = mod.get_function('step_d')
        jitfunc2 = mod.get_function('add_sources_d')
        fd1_d = mod.get_global('fd1_d')[0]
        fd2_d = mod.get_global('fd2_d')[0]

        pad_width = 2
        super(Scalar1D, self).__init__(jitfunc1, jitfunc2, fd1_d, fd2_d,
                                       model, dx, source_dt, sources,
                                       pad_width, pml_width=pml_width)

    def step(self, num_steps=1):

        for it in range(num_steps):
            for inner_it in range(self.timestep.step_ratio):
                self.jitfunc1(self.wavefield.current_gpu,
                              self.wavefield.previous_gpu,
                              self.pml.phi[0].current_gpu,
                              self.pml.phi[0].previous_gpu,
                              self.pml.sigma_gpu[0],
                              self.model.padded_property_gpu['vp2dt2'],
                              np.float32(self.timestep.dt),
                              np.int32(self.geometry.model_shape_padded[0]),
                              grid=self.griddim, block=self.blockdim)
                self.jitfunc2(self.wavefield.previous_gpu,
                              self.model.padded_property_gpu['vp2dt2'],
                              self.sources.amplitude_gpu,
                              self.sources.padded_locations_gpu,
                              np.int32(self.timestep.step_idx),
                              np.int32(self.geometry.model_shape_padded[0]),
                              np.int32(self.timestep.num_steps),
                              np.int32(self.sources.num_sources_per_shot),
                              grid=(self.sources.num_shots, 1),
                              block=(self.sources.num_sources_per_shot, 1, 1))
                self.wavefield.current_gpu, self.wavefield.previous_gpu = \
                    self.wavefield.previous_gpu, self.wavefield.current_gpu
                self.pml.phi[0].current_gpu, self.pml.phi[0].previous_gpu = \
                    self.pml.phi[0].previous_gpu, self.pml.phi[0].current_gpu

            self.timestep.step_idx += 1

        drv.memcpy_dtoh(self.wavefield.current, self.wavefield.current_gpu)

        return self.wavefield.inner



import numpy as np

class RecordReceivers(object):

    def __init__(self, propagator, receiver_locations):
        if receiver_locations is None:
            receiver_locations = np.array([], np.int)
        self.receiver_locations = receiver_locations
        num_steps = propagator.timestep.num_steps
        num_receivers = receiver_locations.shape[:-1]
        self.receivers = np.zeros(num_receivers + (num_steps, ), np.float32)
        self.step = 0

    def record(self, wavefield, step):
        pass



"""Wave equation propagator.
"""
import numpy as np
from smii.modeling.propagators import (libscalar1d, libscalar2d)

class ScalarPropagator(object):
    """A scalar wave equation propagator object."""
    def __init__(self, model, dx, source_dt, sources, pad_width,
                 pml_width=None):
        """A scalar wave equation propagator base class.

        Arguments:
            model: An array containing wave speed
            dx: A float specifying cell size
            source_dt: A float specifying source time step size
            sources: A dictionary containing two keys:
                     'amplitude': [num_shots, num_sources_per_shot, nt]
                                  array
                     'location': [num_shots, num_sources_per_shot, ndim]
                                 array
            pad_width: An int specifying the padding to add around the
                       computational domain
            pml_width: An int specifying the number of cells of padding to be
                       added around the model for the PML
        """

        self.sources = Sources(sources, source_dt)
        self.geometry = Geometry(model.shape, dx, pad_width, pml_width,
                                 self.sources.num_shots,
                                 self.sources.num_sources_per_shot,
                                 self.sources.ndim)
        self.sources.add_padding(self.geometry.total_pad)
        self.timestep = Timestep(np.max(model), dx, self.sources.num_steps,
                                 self.sources.dt)
        self.pml = Pml(self.geometry, np.max(model))
        self.model = Model(self.geometry, self.timestep, vp=model)
        self.wavefield = Wavefield(self.geometry)

    def step(self, num_steps=1):
        raise NotImplementedError

    def _step_setup(self, num_steps):
        sources = {'amplitude': self.sources.step_range(self.timestep.step_idx,
                                                        num_steps),
                   'locations': self.sources.padded_locations}
        return sources

    def _update_pointers(self, num_steps):
        if (num_steps * self.timestep.step_ratio)%2 != 0:
            self.wavefield.current, self.wavefield.previous = \
                    self.wavefield.previous, self.wavefield.current
            for phi in self.pml.phi:
                phi.current, phi.previous = phi.previous, phi.current


class Sources(object):
    def __init__(self, sources, source_dt):
        self.amplitude = sources['amplitude'].astype(np.float32)
        self.locations = sources['locations'].astype(np.int32)
        self.dt = source_dt
        self.num_shots = self.amplitude.shape[0]
        self.num_sources_per_shot = self.amplitude.shape[1]
        self.num_steps = self.amplitude.shape[2]
        self.ndim = self.locations.shape[2]

    def add_padding(self, total_pad):
        self.padded_locations = self.locations + total_pad

    def step_range(self, start_step, num_steps):
        return self.amplitude[..., start_step : start_step + num_steps]


class Geometry(object):
    def __init__(self, model_shape, dx, pad_width, pml_width, num_shots,
                 num_sources_per_shot, ndim):

        if pml_width is None:
            pml_width = 10

        self.dx = dx
        self.pad_width = pad_width
        self.pml_width = pml_width
        self.total_pad = self.pml_width + self.pad_width
        self.ndim = len(model_shape)
        assert self.ndim == ndim, ('{} != {}, shape of model and number of '
                                   'dimensions in source locations do not '
                                   'match.'.format(self.ndim, ndim))

        # Make a list of shot indices to use when adding sources to the
        # wavefield. [num_shots, num_sources_per_shot]. For 2 shots
        # with 3 sources per shot: [[0, 0, 0], [1, 1, 1]]
        self.shotidx = np.arange(num_shots).reshape(-1, 1)\
                                           .repeat(num_sources_per_shot,
                                                   axis=1)

        self.model_shape = list(model_shape)
        self.model_shape_padded = [d + 2 * self.total_pad
                                   for d in self.model_shape]

        self.propagation_shape = [num_shots] + self.model_shape
        self.propagation_shape_padded = ([num_shots]
                                         + self.model_shape_padded)


class Timestep(object):
    def __init__(self, max_vel, dx, num_steps, source_dt):

        self.num_steps = num_steps

        max_dt = 0.6 * dx / max_vel
        self.step_ratio = int(np.ceil(source_dt / max_dt))
        self.dt = source_dt / self.step_ratio
        self.step_idx = 0


class Pml(object):
    def __init__(self, geometry, max_vel):

        pml_width = geometry.pml_width
        profile = ((np.arange(pml_width)/pml_width)**2
                   * 3 * max_vel * np.log(1000)
                   / (2 * geometry.dx * pml_width))
        self.sigma = self._set_sigma(geometry, profile)
        self.phi = [Wavefield(geometry) for dim in range(geometry.ndim)]

    def _set_sigma(self, geometry, profile):
        total_pad = geometry.total_pad
        pad_width = geometry.pad_width
        sigma = []
        for dim in range(geometry.ndim):
            sigma_dim = np.zeros(geometry.model_shape_padded[dim], np.float32)
            sigma_dim[total_pad-1:pad_width-1:-1] = profile
            sigma_dim[-total_pad:-pad_width] = profile
            sigma_dim[:pad_width] = sigma_dim[pad_width]
            sigma_dim[-pad_width:] = sigma_dim[-pad_width-1]
            sigma.append(sigma_dim)
        return sigma


class Model(object):
    def __init__(self, geometry, timestep, vp):
        self.property = {}
        self.property['vp'] = vp
        self.property['vp2dt2'] = vp**2 * timestep.dt**2
        self.padded_property = {}
        for key, value in self.property.items():
            self.padded_property[key] = np.pad(value, geometry.total_pad,
                                               'edge')


class Wavefield(object):
    def __init__(self, geometry):
        shape = geometry.propagation_shape_padded
        self.current, self.previous = [np.zeros(shape, np.float32),
                                       np.zeros(shape, np.float32)]
        self._inner_slice = [slice(geometry.total_pad, -geometry.total_pad)] \
                * geometry.ndim

    @property
    def inner(self):
        return self.current[[...] + self._inner_slice]


class Scalar1D(ScalarPropagator):
    """1D scalar wave propagator."""
    def __init__(self, model, dx, source_dt, sources, pml_width=None):
        pad_width = 2
        super(Scalar1D, self).__init__(model, dx, source_dt, sources,
                                       pad_width, pml_width=pml_width)

    def step(self, num_steps=1):
        """Propagate wavefield."""

        sources = self._step_setup(num_steps)

        libscalar1d.scalar1d.step(self.wavefield.current.T,
                                  self.wavefield.previous.T,
                                  self.pml.phi[0].current.T,
                                  self.pml.phi[0].previous.T,
                                  self.pml.sigma[0].T,
                                  self.model.padded_property['vp2dt2'].T,
                                  self.geometry.dx, self.timestep.dt,
                                  sources['amplitude'].T,
                                  sources['locations'].T, num_steps,
                                  self.geometry.pml_width,
                                  self.timestep.step_ratio)

        self._update_pointers(num_steps)

        self.timestep.step_idx += num_steps

        return self.wavefield.inner


class Scalar2D(ScalarPropagator):
    """2D scalar wave propagator."""
    def __init__(self, model, dx, source_dt, sources, pml_width=None):
        pad_width = 2
        super(Scalar2D, self).__init__(model, dx, source_dt, sources,
                                       pad_width, pml_width=pml_width)

    def step(self, num_steps=1):
        """Propagate wavefield."""

        sources = self._step_setup(num_steps)

        libscalar2d.scalar2d.step(self.wavefield.current.T,
                                  self.wavefield.previous.T,
                                  self.pml.phi[1].current.T,
                                  self.pml.phi[1].previous.T,
                                  self.pml.phi[0].current.T,
                                  self.pml.phi[0].previous.T,
                                  self.pml.sigma[1].T,
                                  self.pml.sigma[0].T,
                                  self.model.padded_property['vp2dt2'].T,
                                  self.geometry.dx, self.timestep.dt,
                                  sources['amplitude'].T,
                                  sources['locations'].T, num_steps,
                                  self.geometry.pml_width,
                                  self.timestep.step_ratio)

        self._update_pointers(num_steps)

        self.timestep.step_idx += num_steps

        return self.wavefield.inner



import numpy as np
import scipy.optimize
from smii.modeling.propagators.propagators import (Scalar1D, Scalar2D)
from smii.modeling.imaging_condition.zero_lag_xcor import ZeroLagXcor
from smii.modeling.forward_model import forward_model
from smii.modeling.backpropagate import backpropagate

def fwi(model, dataset, dx, dt, propagator, maxiter=1,
        loss_file=None, true_model=None, bounds=None):
    """Update model using FWI to better match dataset.

    Arguments:
        model: An array containing wave speed.
        dataset: An iterable that returns a batch of sources
                 and receivers.
                 Both of these must be dictionaries that contain:
                    'amplitude': A [batch_size, num_per_shot, nt] array
                    'locations': A [batch_size, num_per_shot, ndim] array
        dx: A float specifying cell size.
        dt: A float specifying source time step size.
        propagator: The propagator class to use.
        maxiter: An int specifying the number of optimizer iterations to
                 perform. Optional - default 1.
        loss_file: The path to a file where the data and model cost will be
                   stored each time they are evaluated. Optional.
        true_model: An array containing the true wave speed model,
                    which will be used to calculate the model error for each
                    model considered. Optional.
        bounds: A tuple containing the minimum and maximum allowable wave
                speeds. Optional.

    Returns:
        The final model
    """
    
    if loss_file is not None:
        loss_file = open(loss_file, 'w')
        loss_file.write('data_cost, model_cost\n')

    if bounds is not None:
        bounds = [bounds] * len(model.ravel())
    
    # Optimisation
    opt = scipy.optimize.minimize(costjac,
                                  model.ravel(),
                                  args=(dataset, dx, dt, propagator,
                                        model.shape, loss_file, true_model),
                                  jac=True,
                                  bounds=bounds,
                                  options={'maxiter': maxiter,
                                           'disp': True},
                                  tol=1e-20,
                                  method='TNC')

    if loss_file is not None:
        loss_file.close()
    return opt.x


def costjac(x, dataset, dx, dt, propagator, shape, loss_file=None,
            true_model=None, compute_grad=True, prop_kwargs=None):

    model = x.reshape(shape)
    
    imaging_condition = ZeroLagXcor
    
    jac = np.zeros(model.shape, np.float32)
    total_cost = 0.0

    if prop_kwargs is None:
        prop_kwargs = {'pml_width': 30}
    
    for sources, receivers in dataset:
        # Forward
        prop = propagator(model, dx, dt, sources, **prop_kwargs)
        recorded_receivers, stored_source_wavefield = \
                forward_model(prop, receivers['locations'],
                              record_receivers=True,
                              store_wavefield=compute_grad)
    
        residual = {}
        residual['amplitude'] = \
                recorded_receivers.receivers - receivers['amplitude']
        residual['locations'] = receivers['locations'].copy()
        cost = float(dt * 0.5
                     * np.linalg.norm(residual['amplitude'].ravel())**2)
        total_cost += cost
    
        if compute_grad:
            # Calculate second time derivative of source
            _second_time_derivative(stored_source_wavefield, model, dt)
        
            # Backpropagation
            residual['amplitude'] = residual['amplitude'][..., ::-1]
            prop = propagator(model, dx, dt, residual, **prop_kwargs)
            batch_jac = imaging_condition(prop.geometry.propagation_shape, dt)
            jac += backpropagate(prop, batch_jac,
                                 stored_source_wavefield).sum(axis=0)
    
    jac *= 2 / model**3

    model_cost = 0.0
    if true_model is not None:
        model_cost = np.linalg.norm(model - true_model)
    
    if loss_file is not None:
        loss_file.write('{}, {}\n'.format(total_cost, model_cost))

    return np.float64(total_cost), jac.ravel().astype(np.float64)


def _second_time_derivative(stored_source_wavefield, model, dt):
    """Calculate second time derivative of source."""
    wavefield = stored_source_wavefield.wavefield
    wavefield[:] = np.gradient(np.gradient(wavefield, axis=0), axis=0) / dt**2



#!/usr/bin/env python
MAJOR               = 0
MINOR               = 0
MICRO               = 3
VERSION             = '%d.%d.%d' % (MAJOR, MINOR, MICRO)
FULLVERSION = VERSION

def write_version_py(filename='smii/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'

version = full_version
"""

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION})
    finally:
        a.close()

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)
    config.add_subpackage('smii')
    config.add_data_files(('smii', 'LICENSE'))
    config.add_data_files(('smii', 'README.md'))
    config.get_version('smii/version.py')
    return config

if __name__ == "__main__":
    from numpy.distutils.core import setup
    write_version_py()
    setup(configuration=configuration,
          author='Alan Richardson',
          author_email='alan@ausargeo.com',
          url='https://github.com/ar4/smii',
          name='smii',
          install_requires=['numpy', 'scipy', 'pycuda', 'pytest'])



#!/usr/bin/env python
def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('propagators', parent_package, top_path)
    config.add_extension(name='libscalar1d', sources=['scalar1d.f90'], extra_f90_compile_args=['-Wall', '-Wextra', '-pedantic', '-Werror', '-march=native', '-O3', '-std=f95', '-fopenmp'], extra_link_args=['-fopenmp'])
    config.add_extension(name='libscalar2d', sources=['scalar2d.f90'], extra_f90_compile_args=['-Wall', '-Wextra', '-pedantic', '-Werror', '-march=native', '-O3', '-std=f95', '-fopenmp'], extra_link_args=['-fopenmp'])
    #config.add_extension(name='libscalar1d', sources=['scalar1d.f90'], extra_f90_compile_args=['-Wall', '-Wextra', '-pedantic', '-Werror', '-march=native', '-O0', '-g', '-fbounds-check', '-std=f95', '-fopenmp'], extra_link_args=['-fopenmp'])
    #config.add_extension(name='libscalar2d', sources=['scalar2d.f90'], extra_f90_compile_args=['-Wall', '-Wextra', '-pedantic', '-Werror', '-march=native', '-O0', '-g', '-fbounds-check', '-std=f95', '-fopenmp'], extra_link_args=['-fopenmp'])

    return config

if __name__ == "__main__":
    from numpy.distutils.core import setup
    setup(configuration=configuration)



import numpy as np
from smii.modeling.record_receivers.record_receivers import RecordReceivers

class NearestCell(RecordReceivers):

    def __init__(self, propagator, receiver_locations):
        super(NearestCell, self).__init__(propagator, receiver_locations)
        self.receiver_locations = \
                np.rint(self.receiver_locations).astype(np.int)

    def record(self, wavefield, step=None):
        if step is None:
            step = self.step

        if self.receiver_locations.shape[-1] == 1:
            for shotidx in range(wavefield.shape[0]):
                self.receivers[shotidx, :, step] = \
                        wavefield[shotidx,
                                  self.receiver_locations[shotidx, :, 0]]

        elif self.receiver_locations.shape[-1] == 2:
            for shotidx in range(wavefield.shape[0]):
                self.receivers[shotidx, :, step] = \
                        wavefield[shotidx,
                                  self.receiver_locations[shotidx, :, 0],
                                  self.receiver_locations[shotidx, :, 1]]
        self.step = step + 1
